#![deny(
    missing_docs,
    missing_debug_implementations,
    missing_copy_implementations,
    trivial_casts,
    trivial_numeric_casts,
    unsafe_code,
    unstable_features,
    unused_import_braces,
    unused_qualifications
)]

//!
//! libFenrir is the official rust library implementing the Fenrir protocol

use ::std::{io::Result, net::SocketAddr, result, sync::Arc};
use ::tokio::{net::UdpSocket, runtime, task::JoinHandle};
use trust_dns_client::client::{
    AsyncClient as DnssecClient, ClientHandle as DnssecHandle,
};

mod config;
pub use config::Config;

/// Instance of a fenrir endpoint
#[allow(missing_copy_implementations, missing_debug_implementations)]
pub struct Fenrir {
    /// library Configuration
    cfg: Config,
    /// internal runtime
    rt: ::tokio::runtime::Runtime,
    /// listening udp sockets
    sockets: Vec<(Arc<UdpSocket>, JoinHandle<Result<()>>)>,
    /// DNSSEC resolvers. multiple for failover
    resolvers: Vec<(
        DnssecClient,
        JoinHandle<result::Result<(), ::trust_dns_proto::error::ProtoError>>,
    )>,
    /// Broadcast channel to tell workers to stop working
    stop_working: ::tokio::sync::broadcast::Sender<bool>,
}

// TODO: graceful vs immediate stop

impl Drop for Fenrir {
    fn drop(&mut self) {
        self.stop_sync()
    }
}

impl Fenrir {
    /// Create a new Fenrir endpoint
    pub fn new(config: &Config) -> Result<Self> {
        let mut builder = runtime::Builder::new_multi_thread();
        if let Some(threads) = config.threads {
            builder.worker_threads(threads.get());
        }
        let rt = match builder.thread_name("libFenrir").build() {
            Ok(rt) => rt,
            Err(e) => {
                ::tracing::error!("Can't initialize: {}", e);
                return Err(e);
            }
        };
        // start listening
        let listen_num = config.listen.len();
        let resolvers_num = config.resolvers.len();
        let (sender, _) = ::tokio::sync::broadcast::channel(1);
        let endpoint = Fenrir {
            cfg: config.clone(),
            rt: rt,
            sockets: Vec::with_capacity(listen_num),
            resolvers: Vec::with_capacity(resolvers_num),
            stop_working: sender,
        };
        Ok(endpoint)
    }
    /// Start all workers, listeners
    pub async fn start(&mut self) -> Result<()> {
        if let Err(e) = self.add_sockets().await {
            self.stop().await;
            return Err(e);
        }
        if let Err(e) = self.setup_dnssec().await {
            self.stop().await;
            return Err(e);
        }
        Ok(())
    }
    /// Stop all workers, listeners
    /// asyncronous version for Drop
    fn stop_sync(&mut self) {
        let mut toempty_resolv = Vec::new();
        let mut toempty_socket = Vec::new();
        ::std::mem::swap(&mut self.resolvers, &mut toempty_resolv);
        ::std::mem::swap(&mut self.sockets, &mut toempty_socket);
        self.rt
            .block_on(Self::real_stop(toempty_socket, toempty_resolv));
    }
    /// Stop all workers, listeners
    pub async fn stop(&mut self) {
        // TODO
        let _ = self.stop_working.send(true);
        let mut toempty_resolv = Vec::new();
        let mut toempty_socket = Vec::new();
        ::std::mem::swap(&mut self.resolvers, &mut toempty_resolv);
        ::std::mem::swap(&mut self.sockets, &mut toempty_socket);
        Self::real_stop(toempty_socket, toempty_resolv).await;
    }
    /// actually do the work of stopping resolvers and listeners
    async fn real_stop(
        sockets: Vec<(Arc<UdpSocket>, JoinHandle<Result<()>>)>,
        resolvers: Vec<(
            DnssecClient,
            JoinHandle<
                result::Result<(), ::trust_dns_proto::error::ProtoError>,
            >,
        )>,
    ) {
        for r in resolvers.into_iter() {
            let _ = r.1.await;
        }
        for s in sockets.into_iter() {
            let _ = s.1.await;
        }
    }
    /// Add all UDP sockets found in config
    /// and start listening for packets
    async fn add_sockets(&mut self) -> Result<()> {
        let sockets = self.cfg.listen.iter().map(|s_addr| async {
            //let socket = self.rt.block_on(bind_udp(s_addr.clone()))?;
            let socket = self.rt.spawn(bind_udp(s_addr.clone())).await??;
            Ok(Arc::new(socket))
        });
        let sockets = ::futures::future::join_all(sockets).await;
        for s_res in sockets.into_iter() {
            match s_res {
                Ok(s) => {
                    let stop_working = self.stop_working.subscribe();
                    let join =
                        self.rt.spawn(listen_udp(stop_working, s.clone()));
                    self.sockets.push((s, join));
                }
                Err(e) => {
                    return Err(e);
                }
            }
        }
        Ok(())
    }
    /// Add and initialize the DNSSEC resolvers to the endpoint
    async fn setup_dnssec(&mut self) -> Result<()> {
        // use a TCP connection to the DNS.
        // the records we need are big, will not fit in a UDP packet
        use tokio::net::TcpStream;
        use trust_dns_client::{
            client::AsyncClient, proto::iocompat::AsyncIoTokioAsStd,
            tcp::TcpClientStream,
        };

        for resolver in self.cfg.resolvers.iter() {
            let (stream, sender) = TcpClientStream::<
                AsyncIoTokioAsStd<TcpStream>,
            >::new(resolver.clone());
            let client_tostart = AsyncClient::new(stream, sender, None);
            // await the connection to be established
            let (client, bg) = client_tostart.await?;

            let handle = self.rt.spawn(bg);
            self.resolvers.push((client, handle));
        }
        Ok(())
    }
    /// get the fenrir data for a domain
    pub async fn resolv(&mut self, domain: &str) -> Result<String> {
        use std::str::FromStr;
        use trust_dns_client::rr::{DNSClass, Name, RData, RecordType};

        let fullname = "_fenrir".to_owned() + domain;
        for client in self.resolvers.iter_mut() {
            let query = client.0.query(
                Name::from_str(&fullname).unwrap(),
                DNSClass::IN,
                RecordType::TXT,
            );

            // wait for its response
            let response = query.await?;

            // validate it's what we expected
            match response.answers()[0].data() {
                Some(RData::TXT(text)) => return Ok(text.to_string()),
                _ => {
                    return Err(::std::io::Error::new(
                        ::std::io::ErrorKind::NotFound,
                        "No TXT record found",
                    ))
                }
            }
        }
        Err(::std::io::Error::new(
            ::std::io::ErrorKind::NotConnected,
            "No DNS server usable",
        ))
    }
}

/// Add an async udp listener
async fn bind_udp(sock: SocketAddr) -> Result<UdpSocket> {
    let socket = UdpSocket::bind(sock).await?;
    // PERF: SO_REUSEADDR/REUSEPORT
    Ok(socket)
}

/// Run a dedicated loop to read packets on the listening socket
async fn listen_udp(
    mut stop_working: ::tokio::sync::broadcast::Receiver<bool>,
    socket: Arc<UdpSocket>,
) -> Result<()> {
    // jumbo frames are 9K max
    let mut buffer: [u8; 9000] = [0; 9000];
    loop {
        let (bytes, sock_from) = ::tokio::select! {
            _done = stop_working.recv() => {
                break;
            }
            result = socket.recv_from(&mut buffer) => {
                result?
            }
        };
        work(&buffer[0..bytes], sock_from).await;
    }
    Ok(())
}

/// Read and do stuff with the udp packet
async fn work(_buffer: &[u8], _sock_from: SocketAddr) {
    // Do nothing for now
}
